#!/usr/bin/env python3
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

"""This module contains tools to manage content for Superdesk editor"""

import re
import logging
import uuid
from textwrap import dedent
from collections.abc import MutableSequence

from flask import current_app as app

from draftjs_exporter.html import HTML
from draftjs_exporter.constants import ENTITY_TYPES, INLINE_STYLES, BLOCK_TYPES
from draftjs_exporter.defaults import STYLE_MAP, BLOCK_MAP
from draftjs_exporter.dom import DOM
from .etree import parse_html, to_string


logger = logging.getLogger(__name__)
DUMMY_RE = re.compile(r"</?dummy_tag>")

ANNOTATION = 'ANNOTATION'
MEDIA = 'MEDIA'
TABLE = 'TABLE'

EDITOR_STATE = 'draftjsState'
ENTITY_MAP = 'entityMap'
ENTITY_RANGES = 'entityRanges'
INLINE_STYLE_RANGES = 'inlineStyleRanges'

TAG_STYLE_MAP = {
    'i': INLINE_STYLES.ITALIC,
    'em': INLINE_STYLES.ITALIC,
    'b': INLINE_STYLES.BOLD,
    'strong': INLINE_STYLES.BOLD,
}

TAG_ENTITY_MAP = {
    'a': ENTITY_TYPES.LINK,
}


def get_field_content_state(item, field):
    try:
        return item['fields_meta'][field][EDITOR_STATE][0]
    except (KeyError, AttributeError):
        return None


def set_field_content_state(item, field, content_state):
    item.setdefault('fields_meta', {}).update({field: {EDITOR_STATE: [content_state]}})


class Entity:
    """Abstraction of a DraftJS entity"""

    def __init__(self, ranges, data):
        self.ranges = ranges
        self.data = data

    @property
    def key(self):
        return str(self.ranges['key'])

    @property
    def type(self):
        return self.data['type']

    @property
    def is_mutable(self):
        return self.data['mutability'] == 'MUTABLE'

    @property
    def content(self):
        return self.data['data']

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return "{mutable}{type} entity (key: {key})".format(
            mutable="mutable " if self.is_mutable else "",
            type=self.type,
            key=self.key,
        )


class EntitySequence(MutableSequence):

    def __init__(self, editor, block):
        self._ranges = block.data.setdefault('entityRanges', [])
        self._mapping = editor.content_state.setdefault('entityMap', {})

    def __getitem__(self, key):
        ranges = self._ranges[key]
        data = self._mapping[str(ranges['key'])]
        return Entity(ranges, data)

    def __setitem__(self, key, value):
        if not isinstance(value, Entity):
            raise TypeError("an Entity instance is expected")
        self._ranges[key] = value.ranges
        self._mapping[value.key] = value.data

    def __delitem__(self, key):
        entity_key = str(self._ranges[key]['key'])
        del self._ranges[key]
        del self._mapping[entity_key]

    def __len__(self):
        return len(self._ranges)

    def insert(self, index, value):
        if not isinstance(value, Entity):
            raise TypeError("an Entity instance is expected")
        self._ranges.insert(index, value.ranges)
        self._mapping[value.key] = value.data


class Block:
    """Abstraction of DraftJS block"""

    def __init__(self, editor, data=None, text=None):
        if data is None:
            data = {
                "key": str(uuid.uuid4()),
                "text": text or "",
                "type": "unstyled",
                "depth": 0,
                "inlineStyleRanges": [],
                "entityRanges": [],
                "data": {},
            }
        self.data = data
        self.entities = EntitySequence(editor, self)

    @property
    def depth(self):
        return self.data['depth']

    @property
    def type(self):
        return self.data['type']

    @property
    def text(self):
        return self.data.get('text')

    @property
    def key(self):
        return self.data.get('key')

    def __str__(self):
        return self.text

    def __repr__(self):
        return "{type} block with text {text!r} and {len_entities} entities".format(
            type=self.type, text=self.text, len_entities=len(self.entities))


class EmbedBlock(Block):

    def __init__(self, editor, html):
        """
        :param str html: raw HTML to embed
        """
        super().__init__(editor)
        self.data.update({
            "text": " ",
            "type": "atomic",
        })
        entity_data = {
            "type": "EMBED",
            "mutability": "MUTABLE",
            "data": {
                "data": {
                    "html": html,
                }
            },
        }
        entity = Entity(
            ranges={"offset": 0, "length": 1, "key": editor.get_next_entity_key()},
            data=entity_data,
        )
        self.entities.append(entity)


class UnstyledBlock(Block):
    def __init__(self, editor, text=None):
        super().__init__(editor)
        if text:
            self.data.update({"text": text})


class BlockSequence(MutableSequence):

    def __init__(self, editor):
        self.editor = editor
        self._blocks = editor.content_state.setdefault('blocks', {})

    def __getitem__(self, key):
        return Block(self.editor, self._blocks[key])

    def __setitem__(self, key, value):
        if not isinstance(value, Block):
            raise TypeError("a Block instance is expected")
        self._blocks[key] = value.data

    def __delitem__(self, key):
        del self._blocks[key]

    def __len__(self):
        return len(self._blocks)

    def insert(self, index, value):
        if not isinstance(value, Block):
            raise TypeError("a Block instance is expected")
        self._blocks.insert(index, value.data)


class EditorContent:
    """Base class to manage content from editors"""

    def __init__(self):
        raise RuntimeError("This class must not be instantiated directly, use create()")

    @staticmethod
    def create(item, field):
        """Factory for EditorContent"""
        return Editor3Content(item, field)


class DraftJSHTMLExporter:

    def __init__(self, content_editor):
        self.content_editor = content_editor
        # XXX: we need one exporter per DraftJSHTMLExporter because
        #      self.render_media needs to access "entityMap" from local
        #      instance. If MEDIA rendering changes in the future, exporter
        #      should be global for all DraftJSHTMLExporter instances
        self.exporter = HTML({
            'engine': DOM.LXML,
            'style_map': dict(STYLE_MAP, **{
                INLINE_STYLES.BOLD: 'b',
                INLINE_STYLES.ITALIC: 'i',
                INLINE_STYLES.FALLBACK: self.style_fallback,
            }),
            'entity_decorators': {
                ENTITY_TYPES.LINK: self.render_link,
                ENTITY_TYPES.HORIZONTAL_RULE: lambda props: DOM.create_element('hr'),
                ENTITY_TYPES.EMBED: self.render_embed,
                MEDIA: self.render_media,
                ANNOTATION: self.render_annotation,
                TABLE: self.render_table,
            },
        })

    @property
    def content_state(self):
        return self.content_editor.content_state

    def render(self):
        blocks = self.content_state['blocks']
        if not blocks:
            return ''
        if blocks and blocks[0]['text'].strip() == '' or blocks[-1]['text'].strip() == '':
            # first and last block may be empty due to client constraints, in this case
            # we must discard them during rendering
            content_state = self.content_state.copy()
            content_state['blocks'] = blocks = blocks[:]
            if blocks[0]['text'].strip() == '' and not blocks[0]['entityRanges']:
                del blocks[0]
            if blocks and blocks[-1]['text'].strip() == '' and not blocks[-1]['entityRanges']:
                del blocks[-1]
            html = self.exporter.render(content_state)
        else:
            html = self.exporter.render(self.content_state)
        # see render_media for details
        return DUMMY_RE.sub('', html)

    def render_annotation(self, props):
        return props['children']

    def render_media(self, props):
        media_props = props['media']
        media_type = media_props.get('type', 'picture')
        rendition = media_props['renditions'].get('original') or media_props['renditions']['viewImage']
        alt_text = media_props.get('alt_text') or ''
        desc = media_props.get('description_text')
        if media_type == 'picture':
            embed_type = "Image"
            elt = DOM.create_element('img', {'src': rendition['href'],
                                             'alt': alt_text}, props['children'])
        elif media_type == 'video':
            embed_type = "Video"
            elt = DOM.create_element('video', {'control': 'control',
                                               'src': rendition['href'],
                                               'alt': alt_text,
                                               'width': '100%',
                                               'height': '100%'}, props['children'])
        elif media_type == 'audio':
            embed_type = "Audio"
            elt = DOM.create_element('audio', {'control': 'control',
                                               'src': rendition['href'],
                                               'alt': alt_text,
                                               'width': '100%',
                                               'height': '100%'}, props['children'])
        else:
            logger.error("Invalid or not implemented media type: {media_type}".format(
                media_type=media_type))
            return None

        content = DOM.render(elt)

        if desc:
            content += "<figcaption>{}</figcaption>".format(desc)

        # we need to retrieve the key, there is not straightforward way to do it
        # so we find the key in entityMap with a corresponding value
        embed_key = next(
            k for k, v in self.content_state['entityMap'].items()
            if v['data'].get('media') == props['media'])

        # <dummy_tag> is needed for the comments, because a root node is necessary
        # it will be removed during rendering.
        embed = DOM.parse_html(dedent("""\
            <dummy_tag><!-- EMBED START {embed_type} {{id: "editor_{key}"}} -->
            <figure>{content}</figure>
            <!-- EMBED END {embed_type} {{id: "editor_{key}"}} --></dummy_tag>""").format(
            embed_type=embed_type,
            key=embed_key,
            content=content)
        )

        return embed

    def render_link(self, props):
        if 'url' in props:
            attribs = {'href': props['url']}
        else:
            link_data = props.get('link', {})
            if link_data.get('attachment'):
                attribs = {'data-attachment': link_data['attachment']}
            elif link_data.get('target'):
                attribs = {'href': link_data['href'], 'target': link_data['target']}
            else:
                attribs = {'href': link_data['href']}

        return DOM.create_element('a', attribs, props['children'])

    def render_embed(self, props):
        embed_pre_process = app.config.get('EMBED_PRE_PROCESS')
        if embed_pre_process:
            for callback in embed_pre_process:
                callback(props['data'])
        # we use superdesk.etree.parse_html instead of DOM.parse_html as the later modify the content
        # and we use directly the wrapping <div> returned with "content='html'". This works because
        # we use the lxml engine with DraftJSExporter.
        div = parse_html(props['data']['html'], content='html')
        div.set('class', 'embed-block')
        description = props.get('description')
        if description:
            p = DOM.create_element('p', {'class': 'embed-block__description'}, description)
            DOM.append_child(div, p)

        return div

    def render_table(self, props):
        num_cols = props['data']['numCols']
        num_rows = props['data']['numRows']
        with_header = props['data'].get('withHeader', False)
        cells = props['data']['cells']
        table = DOM.create_element('table')
        if with_header:
            start_row = 1
            thead = DOM.create_element('thead')
            DOM.append_child(table, thead)
            tr = DOM.create_element('tr')
            DOM.append_child(thead, tr)
            for col_idx in range(num_cols):
                th = DOM.create_element('th')
                DOM.append_child(tr, th)
                try:
                    content_state = cells[0][col_idx]
                except IndexError:
                    continue
                content = DOM.parse_html(self.exporter.render(content_state))
                if content.text or len(content):
                    DOM.append_child(th, content)
        else:
            start_row = 0

        if not with_header or num_rows > 1:
            tbody = DOM.create_element('tbody')
            DOM.append_child(table, tbody)

        for row_idx in range(start_row, num_rows):
            tr = DOM.create_element('tr')
            DOM.append_child(tbody, tr)
            for col_idx in range(num_cols):
                td = DOM.create_element('td')
                DOM.append_child(tr, td)
                try:
                    content_state = cells[row_idx][col_idx]
                except IndexError:
                    continue
                content = DOM.parse_html(self.exporter.render(content_state))
                if content.text or len(content):
                    DOM.append_child(td, content)

        return table

    def style_fallback(self, props):
        type_ = props['inline_style_range']['style']
        # we need to use fallback for annotations, has they have suffixes, it's
        # not only "ANNOTATION"
        if type_.startswith('ANNOTATION'):
            attribs = {"annotation-id": type_[11:]}
            return DOM.create_element('span', attribs, props['children'])
        else:
            logger.error("No style renderer for {type_!r}".format(type_=type_))


class Editor3Content(EditorContent):
    """Handle content for Superdesk Editor 3, using Draft.js ContentState

    see https://medium.com/@rajaraodv/how-draft-js-represents-rich-text-data-eeabb5f25cf2 for documentation
    on ContentState.
    """

    editor_version = 3
    HTML_EXPORTER = DraftJSHTMLExporter

    def __init__(self, item, field='body_html', is_html=True):
        """
        :param item: item containing Draft.js ContentState
        :param field: field to manage, can be "body_html", "headline", etc.
        :param is_html: boolean to indicate if the field is html or text field
        """
        self.item = item
        self.field = field
        self.is_html = is_html
        self.content_state = get_field_content_state(item, field)
        if not self.content_state:
            self._create_state_from_html(item.get(field))
        self.blocks = BlockSequence(self)
        self.html_exporter = DraftJSHTMLExporter(self)

    def _create_state_from_html(self, value=None):
        self.content_state = {
            'blocks': [],
            'entityMap': {},
        }

        if not value:
            return

        def create_entity(entity_type, data, mutability='MUTABLE'):
            key = len(self.content_state['entityMap'].keys())
            self.content_state['entityMap'][str(key)] = {
                'type': entity_type,
                'mutability': mutability,
                'data': data,
            }
            return key

        def create_atomic_block(entity_type, data):
            block = self.create_block('atomic', text=" ").data
            entity_key = create_entity(entity_type, data)
            block['entityRanges'] = [{'offset': 0, 'length': 1, 'key': entity_key}]
            block['inlineStyleRanges'] = []
            self.content_state['blocks'].append(block)

        if self.is_html:
            root = parse_html(value, 'html')
            for i, elem in enumerate(root):
                try:
                    block_type = next((key for key, val in BLOCK_MAP.items() if val == elem.tag))
                    depth = 0
                except StopIteration:
                    block_type = None
                    depth = 0
                    if elem.tag == 'figure':
                        try:
                            m = re.search(r'<!-- EMBED START (?:Image|Video) {id: "(.*)"}', str(root[i - 1]).strip())
                            media = self.item['associations'][m.group(1)]
                            create_atomic_block('MEDIA', {'media': media})
                        except (KeyError, IndexError, AttributeError):
                            create_atomic_block('EMBED', {'data': {'html': to_string(elem, method='html')}})
                        continue
                    elif elem.tag in ('ul', 'ol'):
                        pass  # generate block for each li
                    elif elem.text and '<!-- EMBED' in str(elem):
                        continue
                    elif elem.tag == 'table':
                        data = {'numCols': 0, 'numRows': 0, 'withHeader': False, 'cells': {}}
                        for row, tr in enumerate(elem.iter('tr')):
                            data['numRows'] += 1
                            data['cells'][row] = {}
                            if row == 0 and len(tr):
                                data['numCols'] = len(tr)
                                data['withHeader'] = tr[0].tag == 'th'
                            for col, td in enumerate(tr):
                                data['cells'][row][col] = {
                                    'blocks': [
                                        self.create_block('unstyled', text="".join(td.itertext())).data,
                                    ],
                                    'entityMap': {},
                                }
                        create_atomic_block('TABLE', {'data': data})
                        continue
                    elif elem.tag == 'div':
                        html = ''.join([to_string(child, method='html') for child in elem])
                        if html.startswith('<html><head>'):
                            html = re.sub(
                                r'<\/head><\/html>',
                                '',
                                html.replace('<html><head>', '', 1),
                            )
                        create_atomic_block('EMBED', {'data': {'html': html}})
                        continue
                    else:
                        logger.warning('ignore block %s', str(elem.tag))
                        continue
                block_text = elem.text or ""
                inline_style_ranges = []
                entity_ranges = []
                for child in elem:
                    child_text = "".join(child.itertext())

                    if child.tag in TAG_STYLE_MAP:
                        inline_style_ranges.append({
                            'offset': len(block_text),
                            'length': len(child_text),
                            'style': TAG_STYLE_MAP[child.tag],
                        })

                    if child.tag in TAG_ENTITY_MAP:
                        if not child_text:
                            child_text = " "  # must be non-empty
                        if child.tag == 'a':
                            data = {'link': {'href': child.attrib.get('href'), 'target': child.attrib.get('target')}}
                        else:
                            data = {}
                        entity_key = create_entity(TAG_ENTITY_MAP[child.tag], data)
                        entity_ranges.append({'offset': len(block_text), 'length': len(child_text), 'key': entity_key})

                    if child.tag == 'li':
                        child_type = BLOCK_TYPES.UNORDERED_LIST_ITEM if elem.tag == 'ul' \
                            else BLOCK_TYPES.ORDERED_LIST_ITEM
                        block = self.create_block(child_type, text=child_text).data
                        block.update(depth=depth)
                        self.content_state['blocks'].append(block)

                    block_text += child_text

                    if child.tail and child.tail.strip():
                        block_text += child.tail

                if elem.tail and elem.tail.strip():
                    block_text += elem.tail

                if block_type:  # no block type for ul/ol
                    block = self.create_block(block_type, text=block_text).data
                    block['inlineStyleRanges'] = inline_style_ranges
                    block['entityRanges'] = entity_ranges
                    self.content_state['blocks'].append(block)
        else:
            for line in value.split('\n'):
                self.content_state['blocks'].append(self.create_block(BLOCK_TYPES.UNSTYLED, text=line).data)

    @property
    def html(self):
        return self.html_exporter.render()

    @property
    def text(self):
        return '\n'.join([block.text for block in self.blocks])

    def get_next_entity_key(self):
        """Return a non existing key for entityMap"""
        return max((int(k) for k in self.content_state['entityMap'].keys()), default=-1) + 1

    def update_item(self):
        self.item[self.field] = self.html if self.is_html else self.text
        set_field_content_state(self.item, self.field, self.content_state)

    def create_block(self, block_type, *args, **kwargs):
        cls_name = "{}Block".format(block_type.capitalize())
        if cls_name in globals():
            cls = globals()[cls_name]
            return cls(self, *args, **kwargs)
        else:
            block = Block(self, *args, **kwargs)
            block.data['type'] = block_type
            return block

    def set_blocks(self, blocks):
        try:
            data = self.blocks[0].data.get('data')  # store internal data from first block
        except IndexError:
            data = {}
        self.content_state['blocks'] = [getattr(block, 'data', block) for block in blocks]
        self.blocks = BlockSequence(self)
        if not len(self.blocks):
            self.prepend('Unstyled')
        if not self.blocks[0].data.get('data') and data:
            self.blocks[0].data['data'] = data

    def prepend(self, block_type, *args, **kwargs):
        """Shortcut to prepend a block from its type"""
        block = self.create_block(block_type, *args, **kwargs)
        self.prepend_block(block)
        return block

    def prepend_block(self, block):
        # first block may be empty, this is a client constraint. In this case,
        # we prepend our block after it
        if self.blocks and not self.blocks[0].text.strip():
            index = 1
        else:
            index = 0
        self.blocks.insert(index, block)


def _replace_text(content_state, old, new):
    if not old:
        return
    for block in content_state['blocks']:
        if block.get('type') == 'atomic':
            entity = content_state[ENTITY_MAP][str(block[ENTITY_RANGES][0]['key'])]
            if entity['type'] == 'TABLE':
                cells = entity['data']['data']['cells']
                for row in cells.values():
                    for cell in row.values():
                        _replace_text(cell, old, new)
            continue
        if not block.get('text'):
            continue
        end = 0
        while True:
            try:
                index = block['text'].index(old, end)
                end = index + len(old)
                block['text'] = new.join([block['text'][:index], block['text'][end:]])
                for range_field in (ENTITY_RANGES, INLINE_STYLE_RANGES):
                    if block.get(range_field):
                        ranges = []
                        for range_ in block[range_field]:
                            range_end = range_['offset'] + range_['length']
                            if range_['offset'] > end:  # starting after replaced, move it
                                range_['offset'] += len(new) - len(old)
                                ranges.append(range_)
                            elif range_end <= index:  # starting before replaced text, keep it
                                ranges.append(range_)
                            elif range_['offset'] <= index and range_end >= end:  # contain the text, fix length
                                range_['length'] += len(new) - len(old)
                                ranges.append(range_)
                            else:
                                # remove ranges overlapping with replaced text
                                if range_field == ENTITY_RANGES:
                                    content_state['entityMap'].pop(str(range_["key"]))
                                continue
                        block[range_field] = ranges
            except ValueError:
                break


def replace_text(item, field, old, new, is_html=True):
    """Replace all occurences of old replaced with new.

    It won't replace it in atomic blocks and embeds,
    only text blocks, headings, tables, ul/ol.
    """
    editor = Editor3Content(item, field, is_html)
    _replace_text(editor.content_state, old, new)
    editor.update_item()


def filter_blocks(item, field, filter, is_html=True):
    """Filter content blocks for field.

    It will keep only blocks for which filter returns True.
    """
    editor = Editor3Content(item, field, is_html)
    blocks = []
    for block in editor.blocks:
        if filter(block):
            blocks.append(block)
    editor.set_blocks(blocks)
    editor.update_item()



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import logging
import json

from flask import request, current_app as app
from eve.utils import config
from eve.methods.common import serialize_value
from flask_babel import _

from superdesk import privilege, get_resource_service
from superdesk.notification import push_notification
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.users import get_user_from_request
from superdesk.utc import utcnow
from superdesk.errors import SuperdeskApiError
from superdesk.default_schema import DEFAULT_SCHEMA, DEFAULT_EDITOR
from copy import deepcopy

logger = logging.getLogger(__name__)

KEYWORDS_CV = 'keywords'


privilege(name="vocabularies", label="Vocabularies Management",
          description="User can manage vocabularies' contents.")


# TODO(petr): add api to specify vocabulary schema
vocab_schema = {
    'crop_sizes': {
        'width': {'type': 'integer'},
        'height': {'type': 'integer'},
    }
}


class VocabulariesResource(Resource):
    schema = {
        '_id': {
            'type': 'string',
            'required': True,
            'unique': True,
            'regex': '^[a-zA-Z0-9-_]+$'
        },
        'display_name': {
            'type': 'string',
            'required': True
        },
        'description': {
            'type': 'string',
            'required': False
        },
        'helper_text': {
            'type': 'string',
            'maxlength': 120
        },
        'tags': {
            'type': 'list',
            'required': False,
            'schema': {
                'type': 'dict',
            }
        },
        'popup_width': {
            'type': 'integer',
            'nullable': True
        },
        'type': {
            'type': 'string',
            'required': True,
            'allowed': ['manageable', 'unmanageable'],
        },
        'items': {
            'type': 'list',
            'required': True,
            'schema': {
                'type': 'dict',
                'allow_unknown': True,
                'schema': {
                    'name': {'type': 'string', 'required': False, 'nullable': True},
                }
            }
        },
        'selection_type': {
            'type': 'string',
            'allowed': ['single selection', 'multi selection', 'do not show'],
            'nullable': True
        },
        'schema_field': {
            'type': 'string',
            'required': False,
            'nullable': True
        },
        'dependent': {
            'type': 'boolean',
        },
        'service': {
            'type': 'dict',
        },
        'priority': {
            'type': 'integer'
        },
        'unique_field': {
            'type': 'string',
            'required': False,
            'nullable': True
        },
        'schema': {
            'type': 'dict'
        },
        'field_type': {
            'type': 'string',
            'nullable': True,
        },
        'field_options': {
            'type': 'dict',
        },
        'init_version': {
            'type': 'integer',
        },
        'preffered_items': {
            'type': 'boolean',
        },
        'date_shortcuts': {
            'type': 'list',
            'nullable': True,
            'schema': {
                'type': 'dict',
                'schema': {
                    'value': {'type': 'integer', 'required': True},
                    'term': {'type': 'string', 'required': True},
                    'label': {'type': 'string', 'required': True},
                }
            }
        },
        'custom_field_type': {
            'type': 'string',
            'nullable': True,
        },
        'custom_field_config': {
            'type': 'dict',
            'nullable': True,
        },
    }

    soft_delete = True
    item_url = r'regex("[-_\w]+")'
    item_methods = ['GET', 'PATCH', 'DELETE']
    resource_methods = ['GET', 'POST']
    privileges = {'PATCH': 'vocabularies', 'POST': 'vocabularies', 'DELETE': 'vocabularies'}
    mongo_indexes = {'field_type': [('field_type', 1)]}


class VocabulariesService(BaseService):

    system_keys = set(DEFAULT_SCHEMA.keys()).union(set(DEFAULT_EDITOR.keys()))

    def _validate_items(self, update):
        # if we have qcode and not unique_field set, we want it to be qcode
        try:
            update['schema']['qcode']
        except KeyError:
            pass
        else:
            update.setdefault('unique_field', 'qcode')
        unique_field = update.get('unique_field')
        vocabs = {}
        if 'schema' in update and 'items' in update:
            for index, item in enumerate(update['items']):
                for field, desc in update.get('schema', {}).items():
                    if ((desc.get('required', False) or unique_field == field) and (
                            field not in item or not item[field])):
                        msg = 'Required ' + field + ' in item ' + str(index)
                        payload = {'error': {'required_field': 1}, 'params': {'field': field, 'item': index}}
                        raise SuperdeskApiError.badRequestError(message=msg, payload=payload)

                    elif desc.get('link_vocab') and desc.get('link_field'):
                        if not vocabs.get(desc['link_vocab']):
                            linked_vocab = self.find_one(req=None, _id=desc['link_vocab']) or {}

                            vocabs[desc['link_vocab']] = [
                                vocab.get(desc['link_field'])
                                for vocab in linked_vocab.get('items') or []
                            ]

                        if item.get(field) and item[field] not in vocabs[desc['link_vocab']]:
                            msg = '{} "{}={}" not found'.format(
                                desc['link_vocab'],
                                desc['link_field'],
                                item[field]
                            )
                            payload = {
                                'error': {
                                    'required_field': 1,
                                    'params': {
                                        'field': field,
                                        'item': index
                                    }
                                }
                            }
                            raise SuperdeskApiError.badRequestError(
                                message=msg,
                                payload=payload
                            )

    def on_create(self, docs):
        for doc in docs:
            self._validate_items(doc)

            if doc.get('field_type') and doc['_id'] in self.system_keys:
                raise SuperdeskApiError(
                    message='{} is in use'.format(doc['_id']),
                    payload={'_id': {'conflict': 1}})

            if self.find_one(req=None, **{'_id': doc['_id'], '_deleted': True}):
                raise SuperdeskApiError(
                    message='{} is used by deleted vocabulary'.format(doc['_id']),
                    payload={'_id': {'deleted': 1}})

    def on_created(self, docs):
        for doc in docs:
            self._send_notification(doc, event='vocabularies:created')

    def on_replace(self, document, original):
        self._validate_items(document)
        document[app.config['LAST_UPDATED']] = utcnow()
        document[app.config['DATE_CREATED']] = original.get(app.config['DATE_CREATED'],
                                                            utcnow()) if original else utcnow()
        logger.info("updating vocabulary item: %s", document["_id"])

    def on_fetched(self, doc):
        """Overriding to filter out inactive vocabularies and pops out 'is_active' property from the response.

        It keeps it when requested for manageable vocabularies.
        """

        if request and hasattr(request, 'args') and request.args.get('where'):
            where_clause = json.loads(request.args.get('where'))
            if where_clause.get('type') == 'manageable':
                return doc

        for item in doc[config.ITEMS]:
            self._filter_inactive_vocabularies(item)
            self._cast_items(item)

    def on_fetched_item(self, doc):
        """
        Overriding to filter out inactive vocabularies and pops out 'is_active' property from the response.
        """
        self._filter_inactive_vocabularies(doc)
        self._cast_items(doc)

    def on_update(self, updates, original):
        """Checks the duplicates if a unique field is defined"""
        if 'items' in updates:
            updated = deepcopy(original)
            updated.update(updates)
            self._validate_items(updated)
        unique_field = original.get('unique_field')
        if unique_field:
            self._check_uniqueness(updates.get('items', []), unique_field)

    def on_updated(self, updates, original):
        """
        Overriding this to send notification about the replacement
        """
        self._send_notification(original)

    def on_replaced(self, document, original):
        """
        Overriding this to send notification about the replacement
        """
        self._send_notification(document)

    def on_delete(self, doc):
        """
        Overriding to validate vocabulary deletion
        """
        if 'field_type' not in doc:
            raise SuperdeskApiError.badRequestError('Default vocabularies cannot be deleted')

    def _check_uniqueness(self, items, unique_field):
        """Checks the uniqueness if a unique field is defined

        :param items: list of items to check for uniqueness
        :param unique_field: name of the unique field
        """
        unique_values = []
        for item in items:
            # compare only the active items
            if not item.get('is_active'):
                continue

            if not item.get(unique_field):
                raise SuperdeskApiError.badRequestError("{} cannot be empty".format(unique_field))

            unique_value = str(item.get(unique_field)).upper()

            if unique_value in unique_values:
                raise SuperdeskApiError.badRequestError("Value {} for field {} is not unique".
                                                        format(item.get(unique_field), unique_field))

            unique_values.append(unique_value)

    def _filter_inactive_vocabularies(self, item):
        vocs = item['items']
        active_vocs = ({k: voc[k] for k in voc.keys() if k != 'is_active'}
                       for voc in vocs if voc.get('is_active', True))

        item['items'] = list(active_vocs)

    def _cast_items(self, vocab):
        """Cast values in vocabulary items using predefined schema.

        :param vocab
        """
        schema = vocab_schema.get(vocab.get('_id'), {})
        for item in vocab.get('items', []):
            for field, field_schema in schema.items():
                if field in item:
                    item[field] = serialize_value(field_schema['type'], item[field])

    def _send_notification(self, updated_vocabulary, event='vocabularies:updated'):
        """
        Sends notification about the updated vocabulary to all the connected clients.
        """

        user = get_user_from_request()
        push_notification(event, vocabulary=updated_vocabulary.get('display_name'),
                          user=str(user[config.ID_FIELD]) if user else None,
                          vocabulary_id=updated_vocabulary['_id'])

    def get_rightsinfo(self, item):
        rights_key = item.get('source', item.get('original_source', 'default'))
        all_rights = self.find_one(req=None, _id='rightsinfo')
        if not all_rights or not all_rights.get('items'):
            return {}
        try:
            all_rights['items'] = self.get_locale_vocabulary(all_rights.get('items'), item.get('language'))
            default_rights = next(info for info in all_rights['items'] if info['name'] == 'default')
        except StopIteration:
            default_rights = None
        try:
            rights = next(info for info in all_rights['items'] if info['name'] == rights_key)
        except StopIteration:
            rights = default_rights
        if rights:
            return {
                'copyrightholder': rights.get('copyrightHolder'),
                'copyrightnotice': rights.get('copyrightNotice'),
                'usageterms': rights.get('usageTerms'),
            }
        else:
            return {}

    def get_extra_fields(self):
        return list(self.get(req=None, lookup={'field_type': {'$exists': True, '$ne': None}}))

    def get_custom_vocabularies(self):
        return list(self.get(req=None, lookup={
            'field_type': None,
            'service': {'$exists': True},
        }))

    def get_forbiden_custom_vocabularies(self):
        return list(self.get(req=None, lookup={
            'field_type': None,
            'selection_type': 'do not show',
            'service': {'$exists': True},
        }))

    def get_locale_vocabulary(self, vocabulary, language):
        if not vocabulary or not language:
            return vocabulary
        locale_vocabulary = []
        for item in vocabulary:
            if 'translations' not in item:
                locale_vocabulary.append(item)
                continue
            new_item = item.copy()
            locale_vocabulary.append(new_item)
            for field, values in new_item.get('translations', {}).items():
                if field in new_item and language in values:
                    new_item[field] = values[language]
        return locale_vocabulary

    def add_missing_keywords(self, keywords, language=None):
        if not keywords:
            return
        cv = self.find_one(req=None, _id=KEYWORDS_CV)
        if cv:
            existing = {item['name'].lower() for item in cv.get('items', [])}
            missing = [keyword for keyword in keywords if keyword.lower() not in existing]
            if missing:
                updates = {'items': cv.get('items', [])}
                for keyword in missing:
                    updates['items'].append({
                        'name': keyword,
                        'qcode': keyword,
                        'is_active': True,
                    })
                self.on_update(updates, cv)
                self.system_update(cv['_id'], updates, cv)
                self.on_updated(updates, cv)
        else:
            items = [{
                'name': keyword,
                'qcode': keyword,
                'is_active': True,
            } for keyword in keywords]
            cv = {
                '_id': KEYWORDS_CV,
                'items': items,
                'type': 'manageable',
                'display_name': _('Keywords'),
                'unique_field': 'name',
                'schema': {
                    'name': {},
                    'qcode': {},
                },
            }
            self.post([cv])


def is_related_content(item_name, related_content=None):
    if related_content is None:
        related_content = list(
            get_resource_service('vocabularies').get(req=None, lookup={'field_type': 'related_content'}))

    if related_content and item_name.split('--')[0] in [content['_id'] for content in related_content]:
        return True

    return False



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
from .service import DesksService
from .resource import DesksResource


def init_app(app):
    """Initialize the `desks` API endpoint.

    :param app: the API application object
    :type app: `Eve`
    """
    service = DesksService(datasource='desks', backend=superdesk.get_backend())
    DesksResource(endpoint_name='desks', app=app, service=service)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk import get_resource_service
from os.path import dirname, join
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from apps.auth.db.commands import CreateUserCommand, ImportUsersCommand


fixtures_path = join(dirname(dirname(dirname(dirname(__file__)))), 'tests', 'auth', 'fixtures')


class UsersTestCase(TestCase):

    def test_create_user_command(self):
        if not self.app.config.get('LDAP_SERVER'):
            user = {'username': 'foo', 'password': 'bar', 'email': 'baz', 'password_changed_on': utcnow()}
            cmd = CreateUserCommand()
            cmd.run(user['username'], user['password'], user['email'], admin=True)
            auth_user = get_resource_service('auth_db').authenticate(user)
            self.assertEquals(auth_user['username'], user['username'])

            cmd.run(user['username'], user['password'], user['email'], admin=True)
            auth_user2 = get_resource_service('auth_db').authenticate(user)
            self.assertEquals(auth_user2['username'], user['username'])
            self.assertEquals(auth_user2['_id'], auth_user['_id'])

    def test_create_user_command_no_update(self):
        if not self.app.config.get('LDAP_SERVER'):
            user = {'username': 'foo', 'password': 'bar', 'email': 'baz', 'password_changed_on': utcnow()}
            cmd = CreateUserCommand()
            cmd.run(
                user['username'], user['password'], user['email'], admin=True
            )
            cmd.run(
                user['username'], "new_password", user['email'], admin=True
            )
            get_resource_service('auth_db').authenticate(user)

    def test_import_users(self):
        """users:import is working with JSON files"""
        roles = [
            {'name': 'Writer', 'privileges': {}},
            {'name': 'Admin', 'privileges': {}},
        ]

        self.app.data.insert('roles', roles)

        cmd = ImportUsersCommand()
        print('path', fixtures_path)
        import_file = join(fixtures_path, "import_users_test.json")
        cmd.run(None, import_file)
        users_service = get_resource_service('users')
        roles_services = get_resource_service('roles')

        found_user = users_service.find_one(req=None, username="toto")
        self.assertIsNotNone(found_user)
        role_id = roles_services.find_one(req=None, name="Writer")['_id']
        self.assertEqual(found_user["role"], role_id)

        found_user = users_service.find_one(req=None, username="titi")
        self.assertIsNotNone(found_user)
        role_id = roles_services.find_one(req=None, name="Admin")['_id']
        self.assertEqual(found_user["role"], role_id)

        found_user = users_service.find_one(req=None, username="invalid_no_email")
        self.assertIsNone(found_user)

        found_user = users_service.find_one(req=None, username="invalid_unknown_role")
        self.assertIsNone(found_user)

    def test_import_users_csv(self):
        """users:import is working with CSV files"""
        roles = [
            {'name': 'Writer', 'privileges': {}},
            {'name': 'Admin', 'privileges': {}},
        ]

        self.app.data.insert('roles', roles)

        cmd = ImportUsersCommand()
        import_file = join(fixtures_path, "import_users_test.csv")
        cmd.run(["username", "first_name", "last_name", "sign_off", "email", "role"], import_file)
        users_service = get_resource_service('users')
        roles_services = get_resource_service('roles')

        found_user = users_service.find_one(req=None, username="mr_x")
        self.assertIsNotNone(found_user)
        role_id = roles_services.find_one(req=None, name="Writer")['_id']
        self.assertEqual(found_user["role"], role_id)

        found_user = users_service.find_one(req=None, username="ms_x")
        self.assertIsNotNone(found_user)
        role_id = roles_services.find_one(req=None, name="Admin")['_id']
        self.assertEqual(found_user["role"], role_id)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2016 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from apps.keywords.alchemy import AlchemyKeywordsProvider
from superdesk import get_backend

from apps.suggestions.resource import SuggestionsResource
from apps.suggestions.service import SuggestionsService


def init_app(app):
    endpoint_name = 'suggestions'
    service = SuggestionsService(endpoint_name, backend=get_backend())
    if app.config.get('KEYWORDS_PROVIDER') == 'Alchemy':
        service.provider = AlchemyKeywordsProvider()
    SuggestionsResource(endpoint_name, app=app, service=service)
